# -*- coding: utf-8 -*-


import warnings
import itertools as itt


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as mlines


import Polygon as plg
import datastock as ds


from ..geom import _comp_solidangles


__all__ = ['compute_etendue_los']


# ###############################################################
# ###############################################################
#                       Main
# ###############################################################


def compute_etendue_los(
    coll=None,
    key=None,
    # parameters
    analytical=None,
    numerical=None,
    res=None,
    margin_par=None,
    margin_perp=None,
    # options
    add_points=None,
    # spectro-only
    ind_ap_lim_spectral=None,
    rocking_curve_fw=None,
    rocking_curve_max=None,
    # bool
    convex=None,
    check=None,
    verb=None,
    plot=None,
    store=None,
    overwrite=None,
    debug=None,
):

    # ------------
    # check inputs
    # ------------

    (
        key,
        spectro,
        is2d,
        doptics,
        dcompute,
        analytical,
        numerical,
        res,
        margin_par,
        margin_perp,
        check,
        verb,
        plot,
        store,
        overwrite,
    ) = _check(
        coll=coll,
        key=key,
        analytical=analytical,
        numerical=numerical,
        res=res,
        margin_par=margin_par,
        margin_perp=margin_perp,
        check=check,
        verb=verb,
        plot=plot,
        store=store,
        overwrite=overwrite,
    )

    # -----------
    # verb

    if verb is True:
        msg = f"\nComputing etendue / los for diag '{key}':"
        print(msg)

    # ---------------
    # prepare optics
    # ---------------

    for key_cam, v0 in dcompute.items():

        # ------------------------
        # get equivalent apertures for all pixels

        (
            pinhole, optics, iref,
            x0, x1, iok,
            px, py, pz,
            cx, cy, cz,
            cents0, cents1,
            centsx, centsy, centsz,
            ap_area, plane_nin,
            spectro, dist_cryst2ap,
        ) = coll.get_diagnostic_equivalent_aperture(
            key=key,
            key_cam=key_cam,
            # inital contour
            add_points=add_points,
            # options
            ind_ap_lim_spectral=ind_ap_lim_spectral,
            convex=convex,
            harmonize=True,
            reshape=False,
            # plot
            plot=False,
            verb=verb,
            store=False,
            return_for_etendue=True,
            debug=debug,
        )

        # ------------------------
        # spectro => rocking curve

        if spectro:

            dmat = coll.dobj['crystal'][optics[iref]]['dmat']
            if rocking_curve_fw is None:
                if dmat.get('drock') is not None:
                    rocking_curve_fw = dmat['drock']['FW']
                else:
                    rocking_curve_fw = np.inf

            if rocking_curve_max is None:
                if dmat.get('drock') is not None:
                    kpow = dmat['drock']['power_ratio']
                    rocking_curve_max = np.max(coll.ddata[kpow]['data'])
                else:
                    rocking_curve_max = 1.

        # ------------------------------------------
        # get distance, area, solid_angle, los

        (
            det_area, distances,
            los_x, los_y, los_z,
            cos_los_det, cos_los_ap,
            solid_angles, solid_angles_rc, res,
        ) = _loop_on_pix(
            coll=coll,
            ldet=v0['ldet'],
            # spectro
            spectro=spectro,
            rocking_curve_fw=rocking_curve_fw,
            rocking_curve_max=rocking_curve_max,
            dist_cryst2ap=dist_cryst2ap,
            # optics
            x0=x0,
            x1=x1,
            px=px,
            py=py,
            pz=pz,
            iok=iok,
            cx=cx,
            cy=cy,
            cz=cz,
            centsx=centsx,
            centsy=centsy,
            centsz=centsz,
            plane_nin=plane_nin,
            ap_area=ap_area,
            res=res,
        )

        # --------------------
        # compute analytically

        shape0 = cx.shape
        if analytical is True:

            etend0 = np.full(tuple(np.r_[3 + (spectro is True), shape0]), np.nan)

            # 0th order
            etend0[0, ...] = ap_area * det_area / distances**2

            # 1st order
            etend0[1, ...] = (
                cos_los_ap * ap_area
                * cos_los_det * det_area / distances**2
            )

            # 2nd order
            etend0[2, ...] = cos_los_ap * ap_area * solid_angles

            # 3rd order for spectro
            if spectro:
                etend0[3, ...] = cos_los_ap * ap_area * solid_angles_rc

        else:
            etend0 = None

        # --------------------
        # compute numerically

        etend1 = None
        if numerical is True:

            if spectro is True:
                etend1 = None
                # etend1 = _compute_etendue_numerical_spectro(
                #     coll=coll,
                #     key_diag=key,
                #     key_cam=key_cam,
                #     is2d=is2d,
                #     doptics=doptics,
                #     los_x=los_x,
                #     los_y=los_y,
                #     los_z=los_z,
                #     margin_par=margin_par,
                #     margin_perp=margin_perp,
                # )
            else:
                etend1 = _compute_etendue_numerical(
                    coll=coll,
                    key_diag=key,
                    key_cam=key_cam,
                    ldeti=v0['ldet'],
                    res=res,
                    los_x=los_x,
                    los_y=los_y,
                    los_z=los_z,
                    margin_par=margin_par,
                    margin_perp=margin_perp,
                    check=check,
                    verb=verb,
                )

        # --------------------
        # optional plotting

        if plot is True:
            _ = _plot_etendues(
                etend0=etend0,
                etend1=etend1,
                res=res,
            )

        # --------
        # reshape

        assert los_x.shape == shape0, (los_x.shape, shape0)

        # --------------------
        # return dict

        dcompute[key_cam].update({
            'analytical': etend0,
            'numerical': etend1,
            'res': res,
            'iref': iref,
            'optics': optics,
            'los_x': los_x,
            'los_y': los_y,
            'los_z': los_z,
            'spectro': spectro,
            'iok': iok,
            'is2d': is2d,
            'cx': cx,
            'cy': cy,
            'cz': cz,
            'x0': x0,
            'x1': x1,
            'cents0': cents0,
            'cents1': cents1,
        })

    # ----------
    # store
    # ----------

    if store is not False:

        for key_cam, v0 in dcompute.items():

            # ref
            ref = coll.dobj['camera'][key_cam]['dgeom']['ref']

            # data
            etendue = v0[store][-1, ...]

            if store == 'analytical':
                etend_type = store
            else:
                etend_type = v0['res'][-1]

            # keys
            ketendue = f'{key}_{key_cam}_etend'
            ddata = {
                ketendue: {
                    'data': etendue,
                    'ref': ref,
                    'dim': 'etendue',
                    'quant': 'etendue',
                    'name': 'etendue',
                    'units': 'm2.sr',
                },
            }

            # check vs overwrite
            if ketendue in coll.ddata.keys():
                if overwrite is True:
                    coll.remove_data(ketendue, propagate=False)
                else:
                    msg = (
                        f"data '{ketendue}' already exists!\n"
                        "Use overwrite=True to force overwriting\n"
                    )
                    raise Exception(msg)

            coll.update(ddata=ddata)
            coll._dobj['diagnostic'][key]['doptics'][key_cam]['etendue'] = ketendue
            coll._dobj['diagnostic'][key]['doptics'][key_cam]['etend_type'] = etend_type

            # spectro
            if spectro and store == 'analytical':
                # data
                etendue = v0[store][-2, ...]

                # keys
                ketendue = f'{key}_{key_cam}_etend0'
                ddata = {
                    ketendue: {
                        'data': etendue,
                        'ref': ref,
                        'dim': 'etendue',
                        'quant': 'etendue',
                        'name': 'etendue',
                        'units': 'm2.sr',
                    },
                }

                # check vs overwrite
                if ketendue in coll.ddata.keys():
                    if overwrite is True:
                        coll.remove_data(ketendue, propagate=False)
                    else:
                        msg = (
                            f"data '{ketendue}' already exists!\n"
                            "Use overwrite=True to force overwriting\n"
                        )
                        raise Exception(msg)

                coll.update(ddata=ddata)
                coll._dobj['diagnostic'][key]['doptics'][key_cam]['etendue0'] = ketendue

    return dcompute, store, overwrite


# ################################################################
# ################################################################
#                       Check
# ################################################################


def _check(
    coll=None,
    key=None,
    analytical=None,
    numerical=None,
    res=None,
    margin_par=None,
    margin_perp=None,
    check=None,
    verb=None,
    plot=None,
    store=None,
    overwrite=None,
):

    # --------
    # key
    # ---------

    # allowed values
    lok = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if any([
            len(v1['optics']) > 0
            for v1 in v0['doptics'].values()
        ])
    ]

    # check
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # -------------
    # spectro, is2d

    spectro = coll.dobj['diagnostic'][key]['spectro']
    is2d = coll.dobj['diagnostic'][key]['is2d']

    # --------
    # doptics

    doptics = coll.dobj['diagnostic'][key]['doptics']
    dcompute = {
        k0: {'compute': len(v0['optics']) > 0}
        for k0, v0 in doptics.items()
    }

    # -------------------------------------------------
    # ldeti: list of individual camera dict (per pixel)
    # -------------------------------------------------

    for k0, v0 in doptics.items():

        dgeom = coll.dobj['camera'][k0]['dgeom']
        lind = [range(ss) for ss in dgeom['shape']]

        cx, cy, cz = coll.get_camera_cents_xyz(key=k0)
        dvect = coll.get_camera_unit_vectors(key=k0)
        outline = dgeom['outline']
        out0 = coll.ddata[outline[0]]['data']
        out1 = coll.ddata[outline[1]]['data']
        is2d = dgeom['nd'] == '2d'
        par = dgeom['parallel']

        dcompute[k0]['ldet'] = [
            {
                'cents_x': cx[ind],
                'cents_y': cy[ind],
                'cents_z': cz[ind],
                'outline_x0': out0,
                'outline_x1': out1,
                'nin_x': dvect['nin_x'] if par else dvect['nin_x'][ind],
                'nin_y': dvect['nin_y'] if par else dvect['nin_y'][ind],
                'nin_z': dvect['nin_z'] if par else dvect['nin_z'][ind],
                'e0_x': dvect['e0_x'] if par else dvect['e0_x'][ind],
                'e0_y': dvect['e0_y'] if par else dvect['e0_y'][ind],
                'e0_z': dvect['e0_z'] if par else dvect['e0_z'][ind],
                'e1_x': dvect['e1_x'] if par else dvect['e1_x'][ind],
                'e1_y': dvect['e1_y'] if par else dvect['e1_y'][ind],
                'e1_z': dvect['e1_z'] if par else dvect['e1_z'][ind],
            }
            for ind in itt.product(*lind)
        ]

    # -----------
    # analytical
    # -----------

    analytical = ds._generic_check._check_var(
        analytical, 'analytical',
        types=bool,
        default=True,
    )

    # -----------
    # numerical
    # -----------

    numerical = ds._generic_check._check_var(
        numerical, 'numerical',
        types=bool,
        default=False if spectro else True,
        allowed=[False] if spectro else [True, False],
    )

    # -----------
    # res
    # -----------

    if res is not None:
        res = np.atleast_1d(res).ravel()

    # -----------
    # margin_par
    # -----------

    margin_par = ds._generic_check._check_var(
        margin_par, 'margin_par',
        types=float,
        default=0.05,
    )

    # -----------
    # margin_perp
    # -----------

    margin_perp = ds._generic_check._check_var(
        margin_perp, 'margin_perp',
        types=float,
        default=0.05,
    )

    # -----------
    # check
    # -----------

    check = ds._generic_check._check_var(
        check, 'check',
        types=bool,
        default=True,
    )

    # -----------
    # verb
    # -----------

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # -----------
    # plot
    # -----------

    if plot is None:
        plot = True
    if not isinstance(plot, bool):
        msg = "Arg plot must be a bool"
        raise Exception(msg)

    # -----------
    # store
    # -----------

    lok = [False]
    if analytical is True:
        lok.append('analytical')
    if numerical is True:
        lok.append('numerical')

    store = ds._generic_check._check_var(
        store, 'store',
        default=lok[-1],
        allowed=lok,
    )

    # -----------
    # overwrite
    # -----------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    return (
        key,
        spectro,
        is2d,
        doptics,
        dcompute,
        analytical,
        numerical,
        res,
        margin_par,
        margin_perp,
        check,
        verb,
        plot,
        store,
        overwrite,
    )


# ###############################################################
# ###############################################################
#                    Loop on camera pixels
# ###############################################################


def _loop_on_pix(
    coll=None,
    # detectors
    ldet=None,
    # spectro
    spectro=None,
    rocking_curve_fw=None,
    rocking_curve_max=None,
    dist_cryst2ap=None,
    # equivalent aperture
    x0=None,
    x1=None,
    px=None,
    py=None,
    pz=None,
    iok=None,
    cx=None,
    cy=None,
    cz=None,
    centsx=None,
    centsy=None,
    centsz=None,
    plane_nin=None,
    ap_area=None,
    # extra
    res=None,
):

    # -------------
    # prepare data
    # -------------

    shape0 = cx.shape
    los_x = np.full(shape0, np.nan)
    los_y = np.full(shape0, np.nan)
    los_z = np.full(shape0, np.nan)
    solid_angles = np.zeros(shape0, dtype=float)
    solid_angles_rc = np.zeros(shape0, dtype=float)
    cos_los_det = np.full(shape0, np.nan)
    distances = np.full(shape0, np.nan)
    mindiff = np.full(shape0, np.nan)

    # ------------------------------
    # compute area, solid angle, los
    # ------------------------------

    lind = [range(ss) for ss in shape0]
    for ii, ind in enumerate(itt.product(*lind)):

        if not iok[ind]:
            continue

        # ------------
        # solid angles

        solid_angles[ind] = _comp_solidangles.calc_solidangle_apertures(
            # observation points
            pts_x=centsx[ind],
            pts_y=centsy[ind],
            pts_z=centsz[ind],
            # polygons
            apertures=None,
            detectors=ldet[ii],
            # possible obstacles
            config=None,
            # parameters
            visibility=False,
            return_vector=False,
            # timing
            timing=False,
        )[0, 0]

        # -------------
        # rocking curve

        if spectro:

            out0 = ldet[ii]['outline_x0']
            width0 = np.max(np.abs(np.diff(out0)))
            dist = np.sqrt(
                (centsx[ind] - ldet[ii]['cents_x'])**2
                + (centsy[ind] - ldet[ii]['cents_y'])**2
                + (centsz[ind] - ldet[ii]['cents_z'])**2
            )

            withrc = dist * rocking_curve_fw * (dist + dist_cryst2ap / dist)
            out0_norm = out0 / width0
            ldet[ii]['outline_x0'] = out0_norm * min(width0, withrc)

            solid_angles_rc[ind] = _comp_solidangles.calc_solidangle_apertures(
                # observation points
                pts_x=centsx[ind],
                pts_y=centsy[ind],
                pts_z=centsz[ind],
                # polygons
                apertures=None,
                detectors=ldet[ii],
                # possible obstacles
                config=None,
                # parameters
                visibility=False,
                return_vector=False,
                # timing
                timing=False,
            )[0, 0]

    # -------------
    # rocking curve
    # -------------

    if spectro:
        solid_angles_rc *= rocking_curve_max
    else:
        solid_angles_rc = None

    # -------------
    # normalize los
    # -------------

    los_x = centsx - cx
    los_y = centsy - cy
    los_z = centsz - cz

    distances = np.sqrt(los_x**2 + los_y**2 + los_z**2)

    los_x = los_x / distances
    los_y = los_y / distances
    los_z = los_z / distances

    # ------
    # angles
    # ------

    for ii, ind in enumerate(itt.product(*lind)):
        cos_los_det[ind] = (
            los_x[ind] * ldet[ii]['nin_x']
            + los_y[ind] * ldet[ii]['nin_y']
            + los_z[ind] * ldet[ii]['nin_z']
        )

    # abs() because for spectro nin is the other way around
    cos_los_ap = np.abs(
        los_x * plane_nin[0]
        + los_y * plane_nin[1]
        + los_z * plane_nin[2]
    )


    # -----------
    # surfaces
    # -----------

    # det
    if ldet[0].get('pix_area') is None:
        det_area = plg.Polygon(np.array([
            ldet[0]['outline_x0'],
            ldet[0]['outline_x1'],
        ]).T).area()

    else:
        det_area = ldet[0]['pix_area']

    # -------------------------------------
    # det outline discretization resolution
    # -------------------------------------

    if res is None:

        res = np.sqrt(det_area)
        if np.any(np.isfinite(mindiff)):
            res = min(res, np.nanmin(mindiff))
        if np.any(ap_area > 0.):
            res = min(res, np.sqrt(np.min(ap_area[ap_area > 0.])))

        res = res * np.r_[1., 0.5, 0.1]

    iok = np.isfinite(res)
    iok[iok] = res[iok] > 0
    if not np.any(iok):
        res = np.r_[0.001]
    else:
        res = res[iok]

    return (
        det_area, distances,
        los_x, los_y, los_z,
        cos_los_det, cos_los_ap,
        solid_angles, solid_angles_rc, res,
    )


# ##################################################################
# ##################################################################
#                   op_post interpolation
# ##################################################################


def _get_lpoly_post(coll=None, lop_post_cls=None, lop_post=None):

    if len(lop_post) == 0:
        return

    lpoly_post = []
    for cc, oo in zip(lop_post_cls, lop_post):
        dgeom = coll.dobj[cc][oo]['dgeom']
        if dgeom['type'] == 'planar':
            p0, p1 = dgeom['outline']
            lpoly_post.append((
                coll.ddata[p0]['data'],
                coll.ddata[p1]['data'],
            ))
        else:
            px, py, pz = dgeom['poly']
            lpoly_post.append((
                coll.ddata[px]['data'],
                coll.ddata[py]['data'],
                coll.ddata[pz]['data'],
            ))

    return lpoly_post


# ##################################################################
# ##################################################################
#           Numerical etendue estimation routine
# ##################################################################


def _compute_etendue_numerical(
    coll=None,
    key_diag=None,
    key_cam=None,
    ldeti=None,
    res=None,
    margin_par=None,
    margin_perp=None,
    los_x=None,
    los_y=None,
    los_z=None,
    check=None,
    verb=None,
):

    # ------------
    # prepare

    # aperture
    lop = coll.dobj['diagnostic'][key_diag]['doptics'][key_cam]['optics']
    aperture = {}
    for k0 in lop:
        px, py, pz = coll.get_optics_poly(key=k0)
        aperture[k0] = {
            'poly_x': px,
            'poly_y': py,
            'poly_z': pz,
            'nin': coll.dobj['aperture'][k0]['dgeom']['nin'],
        }

    # shape0 = det['cents_x'].shape
    nd = len(ldeti)

    ap_ind = np.cumsum([v0['poly_x'].size for v0 in aperture.values()][:-1])

    ap_tot_px = np.concatenate(tuple(
        [v0['poly_x'] for v0 in aperture.values()]
    ))
    ap_tot_py = np.concatenate(tuple(
        [v0['poly_y'] for v0 in aperture.values()]
    ))
    ap_tot_pz = np.concatenate(tuple(
        [v0['poly_z'] for v0 in aperture.values()]
    ))

    # ------------------------------
    # Get plane perpendicular to los

    etendue = np.full((res.size, nd), np.nan)
    for ii in range(nd):

        if verb is True:
            msg = f"Numerical etendue for det {ii+1} / {nd}"
            print(msg)

        if np.isnan(los_x[ii]):
            continue

        # det poly 3d
        det_Px = (
            ldeti[ii]['cents_x']
            + ldeti[ii]['outline_x0']*ldeti[ii]['e0_x']
            + ldeti[ii]['outline_x1']*ldeti[ii]['e1_x']
        )
        det_Py = (
            ldeti[ii]['cents_y']
            + ldeti[ii]['outline_x0']*ldeti[ii]['e0_y']
            + ldeti[ii]['outline_x1']*ldeti[ii]['e1_y']
        )
        det_Pz = (
            ldeti[ii]['cents_z']
            + ldeti[ii]['outline_x0']*ldeti[ii]['e0_z']
            + ldeti[ii]['outline_x1']*ldeti[ii]['e1_z']
        )

        # det to ap vectors
        PA_x = ap_tot_px[:, None] - det_Px[None, :]
        PA_y = ap_tot_py[:, None] - det_Py[None, :]
        PA_z = ap_tot_pz[:, None] - det_Pz[None, :]

        sca1 = PA_x * los_x[ii] + PA_y * los_y[ii] + PA_z * los_z[ii]
        # get length along los
        k_los = (1. + margin_par) * np.max(sca1)

        # get center of plane perpendicular to los
        c_los_x = ldeti[ii]['cents_x'] + k_los * los_x[ii]
        c_los_y = ldeti[ii]['cents_y'] + k_los * los_y[ii]
        c_los_z = ldeti[ii]['cents_z'] + k_los * los_z[ii]

        # get projections of corners on plane perp. to los
        sca0 = (
            (c_los_x - det_Px[None, :]) * los_x[ii]
            + (c_los_y - det_Py[None, :]) * los_y[ii]
            + (c_los_z - det_Pz[None, :]) * los_z[ii]
        )
        k_plane = sca0 / sca1

        # get LOS-specific unit vectors

        e0_xi = (
            los_y[ii] * ldeti[ii]['e1_z'] - los_z[ii] * ldeti[ii]['e1_y']
        )
        e0_yi = (
            los_z[ii] * ldeti[ii]['e1_x'] - los_x[ii] * ldeti[ii]['e1_z']
        )
        e0_zi = (
            los_x[ii] * ldeti[ii]['e1_y'] - los_y[ii] * ldeti[ii]['e1_x']
        )

        e0_normi = np.sqrt(e0_xi**2 + e0_yi**2 + e0_zi**2)
        e0_xi = e0_xi / e0_normi
        e0_yi = e0_yi / e0_normi
        e0_zi = e0_zi / e0_normi

        e1_xi = los_y[ii] * e0_zi - los_z[ii] * e0_yi
        e1_yi = los_z[ii] * e0_xi - los_x[ii] * e0_zi
        e1_zi = los_x[ii] * e0_yi - los_y[ii] * e0_xi

        # get projections on det_e0 and det_e1 in plane

        x0 = np.split(
            ((det_Px[None, :] + k_plane * PA_x) - c_los_x)*e0_xi
            + ((det_Py[None, :] + k_plane * PA_y) - c_los_y)*e0_yi
            + ((det_Pz[None, :] + k_plane * PA_z) - c_los_z)*e0_zi,
            ap_ind,
            axis=0,
        )
        x1 = np.split(
            ((det_Px[None, :] + k_plane * PA_x) - c_los_x)*e1_xi
            + ((det_Py[None, :] + k_plane * PA_y) - c_los_y)*e1_yi
            + ((det_Pz[None, :] + k_plane * PA_z) - c_los_z)*e1_zi,
            ap_ind,
            axis=0,
        )

        x0_min = np.max([np.min(x0s) for x0s in x0])
        x0_max = np.min([np.max(x0s) for x0s in x0])
        x1_min = np.max([np.min(x1s) for x1s in x1])
        x1_max = np.min([np.max(x1s) for x1s in x1])

        w0 = x0_max - x0_min
        w1 = x1_max - x1_min

        min_res = min(2*margin_perp*w0, 2*margin_perp*w1)
        too_large = res >= min_res
        if np.any(too_large):
            msg = (
                f"Minimum etendue resolution for det {ii} / {nd}: {min_res}\n"
                "The following res values may lead to errors:\n"
                f"\t- res values = {res}\n"
                f"\t- too large  = {too_large}\n"
            )
            warnings.warn(msg)

        # -------------------
        # Discretize aperture

        for jj in range(res.size):

            coef = 1. + 2.*margin_perp
            n0 = int(np.ceil(coef*w0 / res[jj]))
            n1 = int(np.ceil(coef*w1 / res[jj]))

            d0 = coef*w0 / n0
            d1 = coef*w1 / n1

            ds = d0 * d1

            pts_0 = np.linspace(
                x0_min - margin_perp*w0,
                x0_max + margin_perp*w0,
                n0 + 1,
            )
            pts_1 = np.linspace(
                x1_min - margin_perp*w1,
                x1_max + margin_perp*w1,
                n1 + 1,
            )
            pts_0 = 0.5 * (pts_0[1:] + pts_0[:-1])
            pts_1 = 0.5 * (pts_1[1:] + pts_1[:-1])

            # debug
            # n0, n1 = 2, 2
            # pts_0 = np.r_[pts_0[0], pts_0[0]]
            # pts_1 = np.r_[0, 0]

            pts_x = (
                c_los_x + pts_0[:, None] * e0_xi + pts_1[None, :] * e1_xi
            ).ravel()
            pts_y = (
                c_los_y + pts_0[:, None] * e0_yi + pts_1[None, :] * e1_yi
            ).ravel()
            pts_z = (
                c_los_z + pts_0[:, None] * e0_zi + pts_1[None, :] * e1_zi
            ).ravel()

            if verb is True:
                msg = (
                    f"\tres = {res[jj]} ({jj+1} / {res.size})"
                    f"    nb. of points = {pts_x.size}"
                )
                print(msg)

            # ----------------------------------
            # compute solid angle for each pixel

            if check is True:
                solid_angle = _comp_solidangles.calc_solidangle_apertures(
                    # observation points
                    pts_x=pts_x,
                    pts_y=pts_y,
                    pts_z=pts_z,
                    # polygons
                    apertures=aperture,
                    detectors=ldeti[ii],
                    # possible obstacles
                    config=None,
                    # parameters
                    visibility=False,
                    return_vector=False,
                    return_flat_pts=True,
                    return_flat_det=True,
                )

                sar = solid_angle.reshape((n0, n1))
                c0 = (
                    ((pts_0[0] < x0_min) == np.all(sar[0, :] == 0))
                    and ((pts_0[-1] > x0_max) == np.all(sar[-1, :] == 0))
                    and ((pts_1[0] < x1_min) == np.all(sar[:, 0] == 0))
                    and ((pts_1[-1] > x1_max) == np.all(sar[:, -1] == 0))
                )
                if not c0 and not too_large[jj]:
                    # debug
                    plt.figure()
                    plt.imshow(
                        sar.T,
                        extent=(
                            x0_min - margin_perp*w0, x0_max + margin_perp*w0,
                            x1_min - margin_perp*w1, x1_max + margin_perp*w1,
                        ),
                        interpolation='nearest',
                        origin='lower',
                        aspect='equal',
                    )

                    lc = ['r', 'm', 'c', 'y']
                    for ss in range(len(x0)):
                        iss = np.r_[np.arange(0, x0[ss].shape[0]), 0]
                        plt.plot(
                            x0[ss][iss, :],
                            x1[ss][iss, :],
                            c=lc[ss % len(lc)],
                            marker='o',
                            ls='-',
                        )

                    plt.plot(
                        pts_0, np.mean(pts_1)*np.ones((n0,)),
                        c='k', marker='.', ls='None',
                    )
                    plt.plot(
                        np.mean(pts_0)*np.ones((n1,)), pts_1,
                        c='k', marker='.', ls='None',
                    )
                    plt.gca().set_xlabel('x0')
                    plt.gca().set_xlabel('x1')
                    # import pdb; pdb.set_trace()
                    msg = "Something is wrong with solid_angle or sampling"
                    raise Exception(msg)
                else:
                    etendue[jj, ii] = np.sum(solid_angle) * ds

            else:
                etendue[jj, ii] = _comp_solidangles.calc_solidangle_apertures(
                    # observation points
                    pts_x=pts_x,
                    pts_y=pts_y,
                    pts_z=pts_z,
                    # polygons
                    apertures=aperture,
                    detectors=ldeti[ii],
                    # possible obstacles
                    config=None,
                    # parameters
                    summed=True,
                    visibility=False,
                    return_vector=False,
                    return_flat_pts=True,
                    return_flat_det=True,
                ) * ds

    return etendue

# ##################################################################
# ##################################################################
#           Numerical etendue estimation routine
# ##################################################################


def _compute_etendue_numerical_spectro(
    coll=None,
    key_diag=None,
    key_cam=None,
    is2d=None,
    doptics=None,
    los_x=None,
    los_y=None,
    los_z=None,
    margin_par=None,
):

    # --------------
    # check

    if is2d is False:
        msg = "Can only handle 2d camera spectrometers"
        raise Exception(msg)

    # --------------
    # Get main LOS

    los = np.r_[np.nanmean(los_x), np.nanmean(los_y), np.nanmean(los_z)]
    los_n = los / np.linalg.norm(los)

    e0_cam = coll.dobj['camera'][key_cam]['dgeom']['e0']
    e1_cam = coll.dobj['camera'][key_cam]['dgeom']['e1']

    e0 = np.cross(e1_cam, los_n)
    e0 = e0 / np.linalg.norm(e0)

    e1 = np.cross(los_n, e0)
    e1 = e1 / np.linalg.norm(e1)

    # get length along los
    k_los = (1. + margin_par) * np.max(sca1)

    # -------------------------
    # grid perpendicular to los

    # pojections of all los on plane
    sca0 = (
        (c_los_x - det_Px[None, :]) * los_x[ii]
        + (c_los_y - det_Py[None, :]) * los_y[ii]
        + (c_los_z - det_Pz[None, :]) * los_z[ii]
    )
    k_plane = sca0 / sca1

    # perpendicular limits
    x0 = np.split(
        ((det_Px[None, :] + k_plane * PA_x) - c_los_x)*e0_xi
        + ((det_Py[None, :] + k_plane * PA_y) - c_los_y)*e0_yi
        + ((det_Pz[None, :] + k_plane * PA_z) - c_los_z)*e0_zi,
        ap_ind,
        axis=0,
    )
    x1 = np.split(
        ((det_Px[None, :] + k_plane * PA_x) - c_los_x)*e1_xi
        + ((det_Py[None, :] + k_plane * PA_y) - c_los_y)*e1_yi
        + ((det_Pz[None, :] + k_plane * PA_z) - c_los_z)*e1_zi,
        ap_ind,
        axis=0,
    )

    x0_min = np.max([np.min(x0s) for x0s in x0])
    x0_max = np.min([np.max(x0s) for x0s in x0])
    x1_min = np.max([np.min(x1s) for x1s in x1])
    x1_max = np.min([np.max(x1s) for x1s in x1])

    w0 = x0_max - x0_min
    w1 = x1_max - x1_min

    min_res = min(2*margin_perp*w0, 2*margin_perp*w1)
    too_large = res >= min_res
    if np.any(too_large):
        msg = (
            f"Minimum etendue resolution for det {ii} / {nd}: {min_res}\n"
            "The following res values may lead to errors:\n"
            f"\t- res values = {res}\n"
            f"\t- too large  = {too_large}\n"
        )
        warnings.warn(msg)

    # 2d grid perpendicular to los
    n0 = int(np.ceil((lim0[1] -lim0[0]) / res))
    n1 = int(np.ceil((lim1[1] -lim1[0]) / res))
    x0 = np.linspace(lim0[0], lim0[1], n0)
    x1 = np.linspace(lim1[0], lim1[1], n1)

    x0f = np.repeat(x0[:, None], n1, axis=1)
    x1f = np.repeat(x1[None, :], n0, axis=0)

    # ------------
    # loop on pts

    import pdb; pdb.set_trace()     # DB



    return etendue


# ##################################################################
# ##################################################################
#                   Plotting routine
# ##################################################################


def _plot_etendues(
    etend0=None,
    etend1=None,
    res=None,
):

    # -------------
    # prepare data

    nmax = 0
    if etend0 is not None:
        if etend0.ndim > 2:
            etend0 = etend0.reshape((etend0.shape[0], -1))
        nmax = max(nmax, etend0.shape[0])
    if etend1 is not None:
        if etend1.ndim > 2:
            etend1 = etend1.reshape((etend1.shape[0], -1))
        nmax = max(nmax, etend1.shape[0])

    x0 = None
    if etend0 is not None:
        x0 = [
            f'order {ii}' for ii in range(nmax)
        ]
    if etend1 is not None:
        x1 = [f'{res[ii]}' for ii in range(nmax)]
        if x0 is None:
            x0 = x1
        else:
            x0 = [f'{x0[ii]}\n{x1[ii]}' for ii in range(nmax)]

    # -------------
    # prepare axes

    fig = plt.figure(figsize=(11, 6))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    ax.set_ylabel('Etendue ' + r'($m^2.sr$)', size=12, fontweight='bold')
    ax.set_xlabel('order of approximation', size=12, fontweight='bold')

    ax.set_xticks(range(0, nmax))
    ax.set_xticklabels(x0)

    # -------------
    # plot

    if etend0 is not None:
        lines = ax.plot(
            etend0,
            ls='-',
            marker='o',
            ms=6,
        )
        lcol = [ll.get_color() for ll in lines]
    else:
        lcol = [None for ii in range(etend1.shape[1])]

    if etend1 is not None:
        for ii in range(etend1.shape[1]):
            ax.plot(
                etend1[:, ii],
                ls='--',
                marker='*',
                ms=6,
                color=lcol[ii],
            )

    # -------------
    # legend

    handles = [
        mlines.Line2D(
            [], [],
            c='k', marker='o', ls='-', ms=6,
            label='analytical',
        ),
        mlines.Line2D(
            [], [],
            c='k', marker='*', ls='--', ms=6,
            label='numerical',
        ),
    ]
    ax.legend(handles=handles)

    return ax


# cython: language_level=3

import os
from .openmp_enabled import is_openmp_enabled

IF TOFU_OPENMP_ENABLED:
    cimport openmp

cpdef get_effective_num_threads(n_threads=None):
    """
    Based on Scikit-learn:  scikit-learn/sklearn/utils/_openmp_helpers.pyx
    Determine the effective number of threads to be used for OpenMP calls
      - For ``n_threads = None``,
        - if the ``OMP_NUM_THREADS`` environment variable is set, return
          ``openmp.omp_get_max_threads()``
        - otherwise return the minimum between ``openmp.omp_get_max_threads()``
          and the number of cpus, taking cgroups quotas into account. Cgroups
          quotas can typically be set by tools such as Docker.
        The result of ``omp_get_max_threads`` can be influenced by environment
        variable ``OMP_NUM_THREADS`` or at runtime by ``omp_set_num_threads``.
      - For ``n_threads > 0``, return this as the maximal number of threads for
        parallel OpenMP calls.
      - For ``n_threads < 0``, return the maximal number of threads minus
        ``|n_threads + 1|``. In particular ``n_threads = -1`` will use as many
        threads as there are available cores on the machine.
      - Raise a ValueError for ``n_threads = 0``.
      If scikit-learn is built without OpenMP support, always return 1.
    """
    if n_threads == 0:
        raise ValueError("n_threads = 0 is invalid")

    local_openmp_enabled = is_openmp_enabled()
    assert local_openmp_enabled == TOFU_OPENMP_ENABLED

    IF TOFU_OPENMP_ENABLED:

        if os.getenv("OMP_NUM_THREADS"):
            # Fall back to user provided number of threads making it possible
            # to exceed the number of cpus.
            max_n_threads = openmp.omp_get_max_threads()
        else:
            max_n_threads = min(openmp.omp_get_max_threads(),
                                len(os.sched_getaffinity(0)))

        if n_threads is None:
            return max_n_threads
        elif n_threads < 0:
            return max(1, max_n_threads + n_threads + 1)

        return min(n_threads, max_n_threads)
    ELSE:
        # OpenMP disabled at build-time => sequential mode
        return 1



""" A tomography library for fusion devices (tokamaks)

See:
https://github.com/ToFuProject/tofu
"""

# Built-in
import os
import glob
import shutil
import logging
import platform
import subprocess
from codecs import open
# ... setup tools
from setuptools import setup, find_packages
# ... for `clean` command
from distutils.command.clean import clean as Clean


# ... packages that need to be in pyproject.toml
import numpy as np
from Cython.Distutils import Extension
from Cython.Distutils import build_ext


# ... local script
import _updateversion as up
# ... openmp utilities
from tofu_helpers.openmp_helpers import is_openmp_installed


# == Checking platform ========================================================
is_platform_windows = platform.system() == "Windows"


# === Setting clean command ===================================================
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("tofu.setup")


class CleanCommand(Clean):

    description = "Remove build artifacts from the source tree"

    def expand(self, path_list):
        """
        Expand a list of path using glob magic.
        :param list[str] path_list: A list of path which may contains magic
        :rtype: list[str]
        :returns: A list of path without magic
        """
        path_list2 = []
        for path in path_list:
            if glob.has_magic(path):
                iterator = glob.iglob(path)
                path_list2.extend(iterator)
            else:
                path_list2.append(path)
        return path_list2

    def find(self, path_list):
        """Find a file pattern if directories.
        Could be done using "**/*.c" but it is only supported in Python 3.5.
        :param list[str] path_list: A list of path which may contains magic
        :rtype: list[str]
        :returns: A list of path without magic
        """
        import fnmatch

        path_list2 = []
        for pattern in path_list:
            for root, _, filenames in os.walk("."):
                for filename in fnmatch.filter(filenames, pattern):
                    path_list2.append(os.path.join(root, filename))
        return path_list2

    def run(self):
        Clean.run(self)

        cython_files = self.find(["*.pyx"])
        cythonized_files = [
            path.replace(".pyx", ".c") for path in cython_files
        ]
        so_files = self.find(["*.so"])
        # really remove the directories
        # and not only if they are empty
        to_remove = [self.build_base]
        to_remove = self.expand(to_remove)
        to_remove += cythonized_files
        to_remove += so_files

        if not self.dry_run:
            for path in to_remove:
                try:
                    if os.path.isdir(path):
                        shutil.rmtree(path)
                    else:
                        os.remove(path)
                    logger.info("removing '%s'", path)
                except OSError:
                    pass
# =============================================================================


# == Getting tofu version =====================================================
_HERE = os.path.abspath(os.path.dirname(__file__))


def get_version_tofu(path=_HERE):

    # Try from git
    isgit = ".git" in os.listdir(path)
    if isgit:
        try:
            git_branch = (
                subprocess.check_output(
                    [
                        "git",
                        "rev-parse",
                        "--abbrev-ref",
                        "HEAD",
                    ]
                )
                .rstrip()
                .decode()
            )
            deploy_branches = ["master", "deploy-test"]
            if (git_branch in deploy_branches or "TRAVIS_TAG" in os.environ):
                version_tofu = up.updateversion()
            else:
                isgit = False
        except Exception:
            isgit = False

    if not isgit:
        version_tofu = os.path.join(path, "tofu")
        version_tofu = os.path.join(version_tofu, "version.py")
        with open(version_tofu, "r") as fh:
            version_tofu = fh.read().strip().split("=")[-1].replace("'", "")

    version_tofu = version_tofu.lower().replace("v", "").replace(" ", "")
    return version_tofu


version_tofu = get_version_tofu(path=_HERE)

print("")
print("Version for setup.py : ", version_tofu)
print("")

# =============================================================================

# =============================================================================
# Get the long description from the README file
# Get the readme file whatever its extension (md vs rst)

_README = [
    ff
    for ff in os.listdir(_HERE)
    if len(ff) <= 10 and ff[:7] == "README."
]
assert len(_README) == 1
_README = _README[0]
with open(os.path.join(_HERE, _README), encoding="utf-8") as f:
    long_description = f.read()
if _README[-3:] == ".md":
    long_description_content_type = "text/markdown"
else:
    long_description_content_type = "text/x-rst"
# =============================================================================


# =============================================================================
#  Compiling files
openmp_installed, openmp_flag = is_openmp_installed()

extra_compile_args = ["-O3", "-Wall", "-fno-wrapv"] + openmp_flag
extra_link_args = [] + openmp_flag

extensions = [
    Extension(
        name="tofu.geom._GG",
        sources=["tofu/geom/_GG.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        language_level="3",
    ),
    Extension(
        name="tofu.geom._basic_geom_tools",
        sources=["tofu/geom/_basic_geom_tools.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._distance_tools",
        sources=["tofu/geom/_distance_tools.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._sampling_tools",
        sources=["tofu/geom/_sampling_tools.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._raytracing_tools",
        sources=["tofu/geom/_raytracing_tools.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._vignetting_tools",
        sources=["tofu/geom/_vignetting_tools.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._chained_list",
        sources=["tofu/geom/_chained_list.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._sorted_set",
        sources=["tofu/geom/_sorted_set.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    ),
    Extension(
        name="tofu.geom._openmp_tools",
        sources=["tofu/geom/_openmp_tools.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        cython_compile_time_env=dict(TOFU_OPENMP_ENABLED=openmp_installed),
    ),
]


setup(
    name="tofu",
    version="{ver}".format(ver=version_tofu),
    # Use scm to get code version from git tags
    # cf. https://pypi.python.org/pypi/setuptools_scm
    # Versions should comply with PEP440. For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    # The version is stored only in the setup.py file and read from it (option
    # 1 in https://packaging.python.org/en/latest/single_source_version.html)
    use_scm_version=False,

    # Description of what tofu does
    description="A python library for Tomography for Fusion",
    long_description=long_description,
    long_description_content_type=long_description_content_type,

    # The project's main homepage.
    url="https://github.com/ToFuProject/tofu",
    # Author details
    author="Didier VEZINET and Laura MENDOZA",
    author_email="didier.vezinet@gmail.com",

    # Choose your license
    license="MIT",

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        # In which language most of the code is written ?
        "Natural Language :: English",
    ],

    # What does your project relate to?
    keywords="tomography geometry 3D inversion synthetic fusion",

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(
        exclude=[
            "doc",
            "_Old",
            "_Old_doc",
            "plugins",
            "plugins.*",
            "*.plugins.*",
            "*.plugins",
            "*.tests10_plugins",
            "*.tests10_plugins.*",
            "tests10_plugins.*",
            "tests10_plugins",
        ]
    ),

    # packages = ['tofu','tofu.geom'],
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    # py_modules=["my_module"],
    # List run-time dependencies here. These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "setuptools>=40.8.0, <64",
        "numpy",
        "scipy",
        # "scikit-sparse",
        # "scikit-umfpack",
        "matplotlib",
        "contourpy",
        "requests",
        "svg.path",
        "Polygon3",
        "cython>=0.26",
        "spectrally>=0.0.7",
    ],
    python_requires=">=3.6",

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": [
            "check-manifest",
            "coverage",
            "pytest",
            "sphinx",
            "sphinx-gallery",
            "sphinx_bootstrap_theme",
        ]
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here. If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #    # If any package contains *.txt, *.rst or *.npz files, include them:
    #    '': ['*.txt', '*.rst', '*.npz'],
    #    # And include any *.csv files found in the 'ITER' package, too:
    #    'ITER': ['*.csv'],
    # },
    package_data={
        "tofu.tests.tests01_geom.test_data": [
            "*.py", "*.txt", ".svg", ".npz"
        ],
        "tofu.tests.tests04_spectro.test_data": ["*.npz"],
        "tofu.tests.tests06_mesh.test_data": ['*.txt', '*.npz'],
        "tofu.geom.inputs": ["*.txt"],
        "tofu.spectro": ["*.txt"],
        "tofu.mag.mag_ripple": ['*.sh', '*.f']
    },
    include_package_data=True,

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html
    # installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # executable scripts can be declared here
    # They can be python or non-python scripts
    # scripts=[
    # ],

    # entry_points point to functions in the package
    # Theye are generally preferable over scripts because they provide
    # cross-platform support and allow pip to create the appropriate form
    # of executable for the target platform.
    entry_points={
        'console_scripts': [
            'tofuplot=tofu.entrypoints.tofuplot:main',
            'tofucalc=tofu.entrypoints.tofucalc:main',
            'tofu-version=scripts.tofuversion:main',
            'tofu-custom=scripts.tofucustom:main',
            'tofu=scripts.tofu_bash:main',
        ],
    },

    py_modules=['_updateversion'],

    # Extensions and commands
    ext_modules=extensions,
    cmdclass={"build_ext": build_ext,
              "clean": CleanCommand},
    include_dirs=[np.get_include()],
)



# -*- coding: utf-8 -*-


# Built-in
import sys
import os
# import itertools as itt
import copy
import warnings
# from abc import ABCMeta, abstractmethod
import inspect
import copy


# Common
import numpy as np
import scipy.interpolate as scpinterp
# import matplotlib.pyplot as plt
# from matplotlib.tri import Triangulation as mplTri
import datastock as ds


# tofu
# from tofu import __version__ as __version__
import tofu.utils as utils
from . import _DataCollection_check_inputs
from . import _comp
from . import _DataCollection_comp
from . import _def
# from . import _comp_spectrallines


__all__ = ['DataCollectionBase']    # , 'TimeTraceCollection']


_INTERPT = 'zero'


#############################################
#############################################
#       Abstract Parent class
#############################################
#############################################


class DataCollection0(utils.ToFuObject):
    """ A generic class for handling data

    Provides methods for:
        - introspection
        - plateaux finding
        - visualization

    """
    # __metaclass__ = ABCMeta

    # Fixed (class-wise) dictionary of default properties
    _ddef = {
        'Id': {'include': ['Mod', 'Cls', 'Name', 'version']},
        'params': {
            'ddata': {
                'units':  (str, 'a.u.'),
                'dim':    (str, 'unknown'),
                'quant':  (str, 'unknown'),
                'name':   (str, 'unknown'),
                'source': (str, 'unknown'),
            },
            'dobj': {},
            'dstatic': {},
         },
    }

    # _dallowed_params = None
    _data_none = None
    _reserved_keys = None

    _show_in_summary_core = ['shape', 'ref']
    _show_in_summary = 'all'
    _max_ndim = None

    _dref = {}
    _dstatic = {}
    _ddata = {}
    _dobj = {}

    def __init_subclass__(cls, **kwdargs):
        # Does not exist before Python 3.6 !!!
        # Python 2
        super().__init_subclass__(**kwdargs)
        # Python 3
        # super().__init_subclass__(**kwdargs)
        cls._ddef = copy.deepcopy(cls._ddef)
        # cls._dplot = copy.deepcopy(Struct._dplot)
        # cls._set_color_ddef(cls._color)

    def __init__(
        self,
        dref=None,
        dstatic=None,
        ddata=None,
        dobj=None,
        Id=None,
        Name=None,
        fromdict=None,
        SavePath=None,
        include=None,
        sep=None,
    ):

        # Create a dplot at instance level
        # self._dplot = copy.deepcopy(self.__class__._dplot)
        kwdargs = locals()
        del kwdargs['self']
        super().__init__(**kwdargs)

    def _reset(self):
        # Run by the parent class __init__()
        super()._reset()
        self._dref = {}
        self._dstatic = {}
        self._ddata = {}
        self._dobj = {}

    @classmethod
    def _checkformat_inputs_Id(cls, Id=None, Name=None,
                               include=None, **kwdargs):
        if Id is not None:
            assert isinstance(Id, utils.ID)
            Name = Id.Name
        # assert isinstance(Name, str), Name
        if include is None:
            include = cls._ddef['Id']['include']
        kwdargs.update({'Name': Name, 'include': include})
        return kwdargs

    ###########
    # Get check and format inputs
    ###########

    ###########
    # _init
    ###########

    def _init(
        self,
        dref=None,
        dstatic=None,
        ddata=None,
        dobj=None,
        **kwargs,
    ):
        self.update(
            dref=dref,
            dstatic=dstatic,
            ddata=ddata,
            dobj=dobj,
        )
        # self._dstrip['strip'] = 0

    ###########
    # set dictionaries
    ###########

    def update(
        self,
        dobj=None,
        ddata=None,
        dref=None,
        dstatic=None,
    ):
        """ Can be used to set/add data/ref

        Will update existing attribute with new dict
        """
        # Check consistency
        (
            self._dref, self._dstatic, self._ddata, self._dobj,
        ) = _DataCollection_check_inputs._consistency(
            dobj=dobj, dobj0=self._dobj,
            ddata=ddata, ddata0=self._ddata,
            dref=dref, dref0=self._dref,
            dstatic=dstatic, dstatic0=self._dstatic,
            reserved_keys=self._reserved_keys,
            ddefparams_data=self._ddef['params']['ddata'],
            ddefparams_obj=self._ddef['params']['dobj'],
            ddefparams_static=self._ddef['params']['dstatic'],
            data_none=self._data_none,
            max_ndim=self._max_ndim,
        )

    # ---------------------
    # Adding ref / quantity one by one
    # ---------------------

    def add_ref(self, key=None, data=None, size=None, **kwdargs):
        dref = {key: {'data': data, 'size': size, **kwdargs}}
        # Check consistency
        self.update(ddata=None, dref=dref, dstatic=None)

    def add_static(self, key=None, which=None, **kwdargs):
        dstatic = {which: {key: kwdargs}}
        # Check consistency
        self.update(ddata=None, dref=None, dstatic=dstatic)

    def add_data(self, key=None, data=None, ref=None, **kwdargs):
        ddata = {key: {'data': data, 'ref': ref, **kwdargs}}
        # Check consistency
        self.update(ddata=ddata, dref=None, dstatic=None)

    def add_obj(self, which=None, key=None, **kwdargs):
        dobj = {which: {key: kwdargs}}
        # Check consistency
        self.update(dobj=dobj, dref=None, dstatic=None)

    # ---------------------
    # Removing ref / quantities
    # ---------------------

    def remove_ref(self, key=None, propagate=None):
        """ Remove a ref (or list of refs) and all associated data """
        (
            self._dref, self._dstatic, self._ddata, self._dobj,
        ) = _DataCollection_check_inputs._remove_ref(
            key=key,
            dref0=self._dref, ddata0=self._ddata,
            dstatic0=self._dstatic,
            dobj0=self._dobj,
            propagate=propagate,
            reserved_keys=self._reserved_keys,
            ddefparams_data=self._ddef['params']['ddata'],
            ddefparams_obj=self._ddef['params']['dobj'],
            data_none=self._data_none,
            max_ndim=self._max_ndim,
        )

    def remove_static(self, key=None, which=None, propagate=None):
        """ Remove a static ref (or list) or a whole category

        key os provided:
            => remove only the desired key(s)
                works only if key is not used in ddata and dobj

        which is provided:
            => treated as param, the whole category of ref_static is removed
                if propagate, the parameter is removed from ddata and dobj
        """
        _DataCollection_check_inputs._remove_static(
            key=key,
            which=which,
            propagate=propagate,
            dstatic0=self._dstatic,
            ddata0=self._ddata,
            dobj0=self._dobj,
        )

    def remove_data(self, key=None, propagate=True):
        """ Remove a data (or list of data) """
        (
            self._dref, self._dstatic, self._ddata, self._dobj,
        ) = _DataCollection_check_inputs._remove_data(
            key=key,
            dref0=self._dref, ddata0=self._ddata,
            dstatic0=self._dstatic,
            dobj0=self._dobj,
            propagate=propagate,
            reserved_keys=self._reserved_keys,
            ddefparams_data=self._ddef['params']['ddata'],
            ddefparams_obj=self._ddef['params']['dobj'],
            data_none=self._data_none,
            max_ndim=self._max_ndim,
        )

    def remove_obj(self, key=None, which=None, propagate=True):
        """ Remove a data (or list of data) """
        (
            self._dref, self._dstatic, self._ddata, self._dobj,
        ) = _DataCollection_check_inputs._remove_obj(
            key=key,
            which=which,
            dobj0=self._dobj,
            ddata0=self._ddata,
            dref0=self._dref,
            dstatic0=self._dstatic,
            reserved_keys=self._reserved_keys,
            ddefparams_data=self._ddef['params']['ddata'],
            ddefparams_obj=self._ddef['params']['dobj'],
            data_none=self._data_none,
            max_ndim=self._max_ndim,
        )

    # ---------------------
    # Get / set / add / remove params
    # ---------------------

    def __check_which(self, which=None, return_dict=None):
        """ Check which in ['data'] + list(self._dobj.keys() """
        return _DataCollection_check_inputs._check_which(
            dref=self._dref,
            ddata=self._ddata,
            dobj=self._dobj,
            dstatic=self._dstatic,
            which=which,
            return_dict=return_dict,
        )

    def get_lparam(self, which=None):
        """ Return the list of params for the chosen dict

        which can be:
            - 'ref'
            - 'data'
            - dobj[<which>]
            - dstatic[<which>]
        """
        which, dd = self.__check_which(which, return_dict=True)
        return list(list(dd.values())[0].keys())

    def get_param(
        self,
        param=None,
        key=None,
        ind=None,
        returnas=None,
        which=None,
    ):
        """ Return the array of the chosen parameter (or list of parameters)

        which can be:
            - 'ref'
            - 'data'
            - dobj[<which>]
            - dstatic[<which>]

        param cen be a str or a list of str

        Can be returned as:
            - dict: {param0: {key0: values0, key1: value1...}, ...}
            - np.ndarray: {param0: np.r_[values0, value1...], ...}

        """
        which, dd = self.__check_which(which, return_dict=True)
        return _DataCollection_check_inputs._get_param(
            dd=dd, dd_name=which,
            param=param, key=key, ind=ind, returnas=returnas,
        )

    def set_param(
        self,
        param=None,
        value=None,
        ind=None,
        key=None,
        which=None,
        distribute=None,
    ):
        """ Set the value of a parameter

        which can be:
            - 'ref'
            - 'data'
            - dobj[<which>]
            - dstatic[<which>]

        value can be:
            - None
            - a unique value (int, float, bool, str, tuple) common to all keys
            - an iterable of vlues (array, list) => one for each key

        A subset of keys can be chosen (ind, key, fed to self.select()) to set
        only the value of some key

        """
        which, dd = self.__check_which(which, return_dict=True)
        _DataCollection_check_inputs._set_param(
            dd=dd, dd_name=which,
            param=param, value=value, ind=ind, key=key,
            distribute=distribute,
        )

    def add_param(
        self,
        param,
        value=None,
        which=None,
    ):
        """ Add a parameter, optionnally also set its value """
        which, dd = self.__check_which(which, return_dict=True)
        _DataCollection_check_inputs._add_param(
            dd=dd,
            dd_name=which,
            param=param,
            value=value,
        )

    def remove_param(
        self,
        param=None,
        which=None,
    ):
        """ Remove a parameter, none by default, all if param = 'all' """
        which, dd = self.__check_which(which, return_dict=True)
        _DataCollection_check_inputs._remove_param(
            dd=dd,
            dd_name=which,
            param=param,
        )

    ###########
    # strip dictionaries
    ###########

    def _strip_ddata(self, strip=0, verb=0):
        pass

    ###########
    # _strip and get/from dict
    ###########

    @classmethod
    def _strip_init(cls):
        cls._dstrip['allowed'] = [0, 1]
        nMax = max(cls._dstrip['allowed'])
        doc = """
                 1: None
                 """
        doc = utils.ToFuObjectBase.strip.__doc__.format(doc, nMax)
        cls.strip.__doc__ = doc

    def strip(self, strip=0, verb=True):
        # super()
        super(DataCollection, self).strip(strip=strip, verb=verb)

    def _strip(self, strip=0, verb=True):
        self._strip_ddata(strip=strip, verb=verb)

    def _to_dict(self):
        dout = {
            'dref': {'dict': self._dref, 'lexcept': None},
            'dstatic': {'dict': self._dstatic, 'lexcept': None},
            'ddata': {'dict': self._ddata, 'lexcept': None},
            'dobj': {'dict': self._dobj, 'lexcept': None},
        }
        return dout

    def _from_dict(self, fd):
        for k0 in ['dref', 'ddata', 'dstatic', 'dobj']:
            if fd.get(k0) is not None:
                getattr(self, '_'+k0).update(**fd[k0])
        self.update()

    ###########
    # properties
    ###########

    @property
    def dref(self):
        """ the dict of references """
        return self._dref

    @property
    def dstatic(self):
        """ the dict of references """
        return self._dstatic

    @property
    def ddata(self):
        """ the dict of data """
        return self._ddata

    @property
    def dobj(self):
        """ the dict of obj """
        return self._dobj

    ###########
    # set and propagate indices for refs
    ###########

    def add_indices_per_ref(self, indices=None, ref=None, distribute=None):

        lparam = self.get_lparam(which='ref')
        if 'indices' not in lparam:
            self.add_param('indices', which='ref')

        self.set_param(
            which='ref',
            param='indices',
            key=ref,
            value=np.array(indices).ravel(),
            distribute=distribute,
        )

    def propagate_indices_per_ref(
        self,
        ref=None,
        lref=None,
        ldata=None,
        param=None,
    ):
        """ Propagate the indices set for a ref to all other lref

        Index propagation is done:
            - ldata = list of len() = 1 + len(lref)
                according to arbitrary (monotonous) data for each ref
            - according to a criterion:
                - 'index': set matching indices (default)
                - param: set matching monotonous quantities depending on ref
        """
        _DataCollection_comp.propagate_indices_per_ref(
            ref=ref,
            lref=lref,
            ldata=ldata,
            dref=self._dref,
            ddata=self._ddata,
            param=param,
            lparam_data=self.get_lparam(which='data')
        )

    ###########
    # extract
    ###########

    def extract(self, keys=None):
        """ Extract some selected data and return as new instance """

        # ----------------
        # check inputs

        if keys is None:
            return
        if isinstance(keys, str):
            keys = [keys]

        keys = ds._generic_check._check_var_iter(
            keys, 'keys',
            types=list,
            allowed=self._ddata.keys(),
        )

        # -----------------------------
        # Get corresponding list of ref

        lref = set([
            k0 for k0, v0 in self._dref.items()
            if any([ss in keys for ss in v0['ldata']])
        ])

        # -------------------
        # Populate with ref

        coll = self.__class__()

        lpar = [
            pp for pp in self.get_lparam(which='ref')
            if pp not in ['ldata', 'ldata_monot', 'ind', 'data']
        ]
        for k0 in lref:
            coll.add_ref(
                key=k0,
                **copy.deepcopy({pp: self._dref[k0][pp] for pp in lpar}),
            )

        # -------------------
        # Populate with data

        lpar = [
            pp for pp in self.get_lparam(which='data')
            if pp not in ['shape', 'monot']
        ]
        for k0 in keys:
            coll.add_data(
                key=k0,
                **copy.deepcopy({pp: self._ddata[k0][pp] for pp in lpar}),
            )

        return coll

    ###########
    # General use methods
    ###########

    def to_DataFrame(self, which=None):
        which, dd = self.__check_which(which, return_dict=True)
        if which is None:
            return
        import pandas as pd
        return pd.DataFrame(dd)

    # ---------------------
    # Key selection methods
    # ---------------------

    def select(self, which=None, log=None, returnas=None, **kwdargs):
        """ Return the indices / keys of data matching criteria

        The selection is done comparing the value of all provided parameters
        The result is a boolean indices array, optionally with the keys list
        It can include:
            - log = 'all': only the data matching all criteria
            - log = 'any': the data matching any criterion

        If log = 'raw', a dict of indices arrays is returned, showing the
        details for each criterion

        """
        which, dd = self.__check_which(which, return_dict=True)
        return _DataCollection_check_inputs._select(
            dd=dd, dd_name=which,
            log=log, returnas=returnas, **kwdargs,
        )

    def _ind_tofrom_key(
        self,
        ind=None,
        key=None,
        returnas=int,
        which=None,
    ):
        """ Return ind from key or key from ind for all data """
        which, dd = self.__check_which(which, return_dict=True)
        return _DataCollection_check_inputs._ind_tofrom_key(
            dd=dd, dd_name=which, ind=ind, key=key,
            returnas=returnas,
        )

    def _get_sort_index(self, which=None, param=None):
        """ Return sorting index of self.ddata dict """

        if param is None:
            return

        if param == 'key':
            ind = np.argsort(list(dd.keys()))
        elif isinstance(param, str):
            ind = np.argsort(
                self.get_param(param, which=which, returnas=np.ndarray)[param]
            )
        else:
            msg = "Arg param must be a valid str\n  Provided: {}".format(param)
            raise Exception(msg)
        return ind

    def sortby(self, param=None, order=None, which=None):
        """ sort the self.ddata dict by desired parameter """

        # --------------
        # Check inputs

        # order
        order = ds._generic_check._check_var(
            order,
            'order',
            types=str,
            default='increasing',
            allowed=['increasing', 'reverse'],
        )

        # which
        which, dd = self.__check_which(which, return_dict=True)

        # --------------
        # sort
        ind = self._get_sort_index(param=param, which=which)
        if ind is None:
            return
        if order == 'reverse':
            ind = ind[::-1]

        lk = list(dd.keys())
        dd = {lk[ii]: dd[lk[ii]] for ii in ind}

        if which == 'data':
            self._ddata = dd
        elif which == 'ref':
            self.dref = dd
        elif which in self._dobj.keys():
            self._dobj[which] = dd
        elif which in self._dstatic.keys():
            self._dstatic[which] = dd

    # ---------------------
    # Methods for getting a subset of the collection
    # ---------------------

    # TBC
    def get_drefddata_as_input(self, key=None, ind=None):
        lk = self._ind_tofrom_key(ind=ind, key=key, returnas=str)
        lkr = [kr for kr in self._dref['lkey']
               if any([kr in self._ddata['dict'][kk]['refs'] for kk in lk])]
        dref = {kr: {'data': self._ddata['dict'][kr]['data']} for kr in lkr}
        lkr = dref.keys()
        ddata = {kk: self._ddata['dict'][kk] for kk in lk if kk not in lkr}
        return dref, ddata

    # TBC
    def get_subset(self, key=None, ind=None, Name=None):
        if key is None and ind is None:
            return self
        else:
            dref, ddata = self.get_drefddata_as_input(key=key, ind=ind)
            if Name is None and self.Id.Name is not None:
                Name = self.Id.Name + '-subset'
            return self.__class__(dref=dref, ddata=ddata, Name=Name)

    # ---------------------
    # Methods for exporting plot collection (subset)
    # ---------------------

    # TBC
    """
    def to_PlotCollection(self, key=None, ind=None, Name=None,
                          dnmax=None, lib='mpl'):
        dref, ddata = self.get_drefddata_as_input(key=key, ind=ind)
        if Name is None and self.Id.Name is not None:
            Name = self.Id.Name + '-plot'
        import tofu.data._core_plot as _core_plot
        if lib == 'mpl':
            cls = _core_plot.DataCollectionPlot_mpl
        else:
            raise NotImplementedError
        obj = cls(dref=dref, ddata=ddata, Name=Name)
        if dnmax is not None:
            obj.set_dnmax(dnmax)
        return obj
    """

    # ---------------------
    # Methods for showing data
    # ---------------------

    def get_summary(
        self,
        show_which=None,
        show=None,
        show_core=None,
        sep='  ',
        line='-',
        just='l',
        table_sep=None,
        verb=True,
        return_=False,
    ):
        """ Summary description of the object content """
        # # Make sure the data is accessible
        # msg = "The data is not accessible because self.strip(2) was used !"
        # assert self._dstrip['strip']<2, msg

        # ------------
        # check inputs

        if show_which is None:
            show_which = ['ref', 'data', 'static', 'obj']

        lcol, lar = [], []

        # -----------------------
        # Build for dref

        if 'ref' in show_which and len(self._dref) > 0:
            lcol.append(['ref key', 'size', 'nb. data', 'nb. data monot.'])
            lar.append([
                [
                    k0,
                    str(self._dref[k0]['size']),
                    len(self._dref[k0]['ldata']),
                    len(self._dref[k0]['ldata_monot']),
                ]
                for k0 in self._dref.keys()
            ])

            lp = self.get_lparam(which='ref')
            if 'indices' in lp:
                lcol[0].append('indices')
                for ii, (k0, v0) in enumerate(self._dref.items()):
                    if self._dref[k0]['indices'] is None:
                        lar[0][ii].append(str(v0['indices']))
                    else:
                        lar[0][ii].append(str(list(v0['indices'])))

            if 'group' in lp:
                lcol[0].append('group')
                for ii, (k0, v0) in enumerate(self._dref.items()):
                    lar[0][ii].append(str(self._dref[k0]['group']))

            if 'inc' in lp:
                lcol[0].append('increment')
                for ii, (k0, v0) in enumerate(self._dref.items()):
                    lar[0][ii].append(str(self._dref[k0]['inc']))

        # -----------------------
        # Build for ddata

        if 'data' in show_which and len(self._ddata) > 0:

            if show_core is None:
                show_core = self._show_in_summary_core
            if isinstance(show_core, str):
                show_core = [show_core]

            lp = self.get_lparam(which='data')
            lkcore = ['shape', 'ref']
            assert all([ss in lp + lkcore for ss in show_core])
            col2 = ['data key'] + show_core

            if show is None:
                show = self._show_in_summary
            if show == 'all':
                col2 += [pp for pp in lp if pp not in col2]
            else:
                if isinstance(show, str):
                    show = [show]
                assert all([ss in lp for ss in show])
                col2 += [pp for pp in show if pp not in col2]
            col2 = [cc for cc in col2 if cc != 'data']

            ar2 = []
            for k0 in self._ddata.keys():
                lu = [k0] + [str(self._ddata[k0].get(cc)) for cc in col2[1:]]
                ar2.append(lu)

            lcol.append(col2)
            lar.append(ar2)

        # -----------------------
        # Build for dstatic

        anystatic = (
            len(self._dstatic) > 0
            and any([
                ss in show_which
                for ss in ['static'] + list(self._dstatic.keys())
            ])
        )
        if anystatic:
            for k0, v0 in self._dstatic.items():
                if 'static' in show_which or k0 in show_which:
                    lk = list(list(v0.values())[0].keys())
                    col = [k0] + [pp for pp in lk]
                    ar = [
                        tuple([k1] + [str(v1[kk]) for kk in lk])
                        for k1, v1 in v0.items()
                    ]
                    lcol.append(col)
                    lar.append(ar)

        # -----------------------
        # Build for dobj

        anyobj = (
            len(self._dobj) > 0
            and any([
                ss in show_which
                for ss in ['obj'] + list(self._dobj.keys())
            ])
        )
        if anyobj:
            for k0, v0 in self._dobj.items():
                if 'obj' in show_which or k0 in show_which:
                    lk = self.get_lparam(which=k0)
                    lk = [
                        kk for kk in lk
                        if 'func' not in kk
                        and 'class' not in kk
                        and kk not in ['handle']
                        and not (k0 == 'axes' and kk == 'bck')
                        and all([
                            not isinstance(v1[kk], dict)
                            for v1 in v0.values()
                        ])
                    ]
                    lcol.append([k0] + [pp for pp in lk])
                    lar.append([
                        tuple([k1] + [str(v1[kk]) for kk in lk])
                        for k1, v1 in v0.items()
                    ])

        return self._get_summary(
            lar,
            lcol,
            sep=sep,
            line=line,
            table_sep=table_sep,
            verb=verb,
            return_=return_,
        )

    def show_all(self):
        self.get_summary(show_which=None)

    def show_data(self):
        self.get_summary(show_which=['ref', 'data'])

    def show_interactive(self):
        self.get_summary(show_which=['axes', 'mobile'])

    # -----------------
    # conversion wavelength - energy - frequency
    # ------------------

    @staticmethod
    def convert_spectral(
        data=None,
        units_in=None, units_out=None,
        returnas=None,
    ):
        """ convert wavelength / energy/ frequency

        Available units:
            wavelength: m, mm, um, nm, A
            energy:     J, eV, keV
            frequency:  Hz, kHz, MHz, GHz

        Can also just return the conversion coef if returnas='coef'
        """
        return _comp_spectrallines.convert_spectral(
            data_in=data, units_in=units_in, units_out=units_out,
            returnas=returnas,
        )

    # -----------------
    # Get common ref
    # ------------------

    def _get_common_ref_data_nearest(
        self,
        lkey=None,
        return_all=None,
    ):
        """ Typically used to get a common (intersection) time vector

        Returns a time vector that contains all time points from all data
        Also return a dict of indices to easily remap each time vector to tall
            such that t[ind] = tall (with nearest approximation)

        """
        return _DataCollection_comp._get_unique_ref_dind(
            dd=self._ddata,
            lkey=lkey,
            return_all=return_all,
        )

    def _get_pts_from_mesh(self, key=None):
        """ Get default pts from a mesh """

        # Check key is relevant
        c0 = (
            key in self._ddata.keys()
            and isinstance(self._ddata[key].get('data'), dict)
            and 'type' in self._ddata[key]['data'].keys()
        )
        if not c0:
            msg = (
                "ddata['{}'] does not exist or is not a mesh".format(key)
            )
            raise Exception(msg)

        if self.ddata[key]['data']['type'] == 'rect':
            if self.ddata[key]['data']['shapeRZ'] == ('R', 'Z'):
                R = np.repeat(self.ddata[key]['data']['R'],
                              self.ddata[key]['data']['nZ'])
                Z = np.tile(self.ddata[key]['data']['Z'],
                            self.ddata[key]['data']['nR'])
            else:
                R = np.tile(self.ddata[key]['data']['R'],
                            self.ddata[key]['data']['nZ'])
                Z = np.repeat(self.ddata[key]['data']['Z'],
                              self.ddata[key]['data']['nR'])
            pts = np.array([R, np.zeros((R.size,)), Z])
        else:
            pts = self.ddata[key]['data']['nodes']
            pts = np.array([
                pts[:, 0], np.zeros((pts.shape[0],)), pts[:, 1],
            ])
        return pts

    # ---------------------
    # Method for interpolation
    # ---------------------

    def _get_finterp(
        self,
        idquant=None, idref1d=None, idref2d=None, idmesh=None,
        idq2dR=None, idq2dPhi=None, idq2dZ=None,
        interp_t=None, interp_space=None,
        fill_value=None, ani=None, Type=None,
    ):

        if interp_t is None:
            interp_t = 'nearest'
        if interp_t != 'nearest':
            msg = "'nearest' is the only time-interpolation method available"
            raise NotImplementedError(msg)

        # Get idmesh
        if idmesh is None:
            if idquant is not None:
                # isotropic
                if idref1d is None:
                    lidmesh = [qq for qq in self._ddata[idquant]['ref']]
                else:
                    lidmesh = [qq for qq in self._ddata[idref2d]['ref']]
            else:
                # anisotropic
                assert idq2dR is not None
                lidmesh = [qq for qq in self._ddata[idq2dR]['ref']]
            assert len(lidmesh) == 1
            idmesh = lidmesh[0]

        # Get common time indices
        if interp_t == 'nearest':
            tall, tbinall, ntall, dind = _DataCollection_comp._get_tcom(
                idquant, idref1d, idref2d, idq2dR,
                dd=self._ddata,
            )

        # Get mesh
        if self._ddata[idmesh]['data']['type'] == 'rect':
            mpltri = None
            trifind = self._ddata[idmesh]['data']['trifind']
        else:
            mpltri = self._ddata[idmesh]['data']['mpltri']
            trifind = mpltri.get_trifinder()

        # # Prepare output

        # Interpolate
        # Note : Maybe consider using scipy.LinearNDInterpolator ?
        if idquant is not None:
            vquant = self._ddata[idquant]['data']
            c0 = (
                self._ddata[idmesh]['data']['type'] == 'quadtri'
                and self._ddata[idmesh]['data']['ntri'] > 1
            )
            if c0:
                vquant = np.repeat(
                    vquant,
                    self._ddata[idmesh]['data']['ntri'],
                    axis=0,
                )
            vr1 = self._ddata[idref1d]['data'] if idref1d is not None else None
            vr2 = self._ddata[idref2d]['data'] if idref2d is not None else None

            # add time dimension if none
            if vquant.ndim == 1:
                vquant = vquant[None, :]
            if vr1.ndim == 1:
                vr1 = vr1[None, :]
            if vr2.ndim == 1:
                vr2 = vr2[None, :]

        else:
            vq2dR = self._ddata[idq2dR]['data']
            vq2dPhi = self._ddata[idq2dPhi]['data']
            vq2dZ = self._ddata[idq2dZ]['data']

            # add time dimension if none
            if vq2dR.ndim == 1:
                vq2dR = vq2dR[None, :]
            if vq2dPhi.ndim == 1:
                vq2dPhi = vq2dPhi[None, :]
            if vq2dZ.ndim == 1:
                vq2dZ = vq2dZ[None, :]

        if interp_space is None:
            interp_space = self._ddata[idmesh]['data']['ftype']

        # get interpolation function
        if ani:
            # Assuming same mesh and time vector for all 3 components
            func = _comp.get_finterp_ani(
                idq2dR, idq2dPhi, idq2dZ,
                interp_t=interp_t,
                interp_space=interp_space,
                fill_value=fill_value,
                idmesh=idmesh, vq2dR=vq2dR,
                vq2dZ=vq2dZ, vq2dPhi=vq2dPhi,
                tall=tall, tbinall=tbinall, ntall=ntall,
                indtq=dind.get(idquant),
                trifind=trifind, Type=Type, mpltri=mpltri,
            )
        else:
            func = _comp.get_finterp_isotropic(
                idquant, idref1d, idref2d,
                vquant=vquant, vr1=vr1, vr2=vr2,
                interp_t=interp_t,
                interp_space=interp_space,
                fill_value=fill_value,
                idmesh=idmesh,
                tall=tall, tbinall=tbinall, ntall=ntall,
                mpltri=mpltri, trifind=trifind,
                indtq=dind.get(idquant),
                indtr1=dind.get(idref1d), indtr2=dind.get(idref2d),
            )

        return func

    # TBC
    def _interp_one_dim(x=None, ind=None, key=None,
                        kind=None, bounds_error=None, fill_value=None):
        """ Return a dict of interpolated data

        Uses scipy.inpterp1d with args:
            - kind, bounds_error, fill_value

        The interpolated data is chosen method select() with args:
            - key, ind

        The interpolation is done against a reference vector x
            - x can be a key to an existing ref
            - x can be user-provided array
                (to properly identify the interpolation dimension)

        Returns:
        --------
        dout:       dict
            dict of interpolated data
        dfail:  dict of failed interpolations, with error messages

        """

        # Check x
        assert x is not None
        if isinstance(x) is str:
            if x not in self.lref:
                msg = "If x is a str, it must be a valid ref!\n"
                msg += "    - x: {}\n".format(x)
                msg += "    - self.lref: {}".format(self.lref)
                raise Exception(msg)
            x = self._ddata[x]['data']
        else:
            try:
                x = np.atleast_1d(x).ravel()
            except Exception:
                msg = (
                    "The reference with which to interpolate, x, should be:\n"
                    + "    - a key to an existing ref\n"
                    + "    - a 1d np.ndarray"
                )
                raise Exception(x)

        # Get keys to interpolate
        lk = self._ind_tofrom_key(ind=ind, key=key, returnas=str)

        # Check provided keys are relevant, and get dim index
        dind, dfail = {}, {}
        for kk in lk:
            dind[kk] = self._ddata[kk]['groups'].index(group)

        # Start loop for interpolation
        dout = {}
        for kk in dout.keys():
            shape = self._ddata['dict'][kk]['shape']

            if not isinstance(self._ddata[kk]['data'], np.ndarray):
                dfail[kk] = "Not a np.ndarray !"
                continue

            kr = self._ddata['dict'][kk]['refs'][dind[kk]]
            vr = self._ddata['dict'][kr]['data']
            data = self._ddata['dict'][kk]['data']
            try:
                if dind[kk] == len(shape) - 1:
                    dout[kk] = scpinterp.interp1d(vr, y,
                                                  kind=kind, axis=-1,
                                                  bounds_error=bounds_error,
                                                  fill_value=fill_value,
                                                  assume_sorted=True)(x)
                else:
                    dout[kk] = scpinterp.interp1d(vr, y,
                                                  kind=kind, axis=dind[kk],
                                                  bounds_error=bounds_error,
                                                  fill_value=fill_value,
                                                  assume_sorted=True)(x)

            except Exception as err:
                dfail[kk] = str(err)
        return dout, dfail

    # ---------------------
    # Method for fitting models in one direction
    # ---------------------

    # TBC
    def _fit_one_dim(ind=None, key=None,
                     Type=None, func=None, **kwdargs):
        """ Return the parameters of a fitted function

        The interpolated data is chosen method select() with args:
            - key, ind

        Returns:
        --------
        dout:       dict
            dict of interpolated data
        dfail:  dict of failed interpolations, with error messages

        """
        # Get keys to interpolate
        lk = self._ind_tofrom_key(ind=ind, key=key, returnas=str)

        # Start model fitting loop on data keys
        dout = {}
        for kk in lk:
            x = None
            axis = None
            dfit = _DataCollection_comp.fit(
                self._ddata['dict'][kk]['data'],
                x=x,
                axis=axis,
                func=func,
                Type=Type,
                **kwdargs,
            )
            dout[kk] = dfit

        return dout

    # ---------------------
    # saving => get rid of function
    # ---------------------

    def save(self, path=None, name=None,
             strip=None, sep=None, deep=True, mode='npz',
             compressed=False, verb=True, return_pfe=False):

        # Remove function mpltri if relevant
        lk = [
            k0 for k0, v0 in self._ddata.items()
            if isinstance(v0['data'], dict)
            and 'mpltri' in v0['data'].keys()
        ]
        for k0 in lk:
            del self._ddata[k0]['data']['mpltri']
        lk = [
            k0 for k0, v0 in self._ddata.items()
            if isinstance(v0['data'], dict)
            and 'trifind' in v0['data'].keys()
        ]
        for k0 in lk:
            del self._ddata[k0]['data']['trifind']

        # call parent method
        return super().save(
            path=path, name=name,
            sep=sep, deep=deep, mode=mode,
            strip=strip, compressed=compressed,
            return_pfe=return_pfe, verb=verb
        )



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import shutil
import itertools as itt

# tofu-specific
import tofu.nist2tofu as tfn

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.nist2tofu.test_03_core'
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_here)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test01_nist(object):

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_search_online_by_wavelengthA(self):

        llambmin = [None, 3.94]
        llambmax = [None, 4.]
        lion = [None, 'H', ['ar', 'W44+']]

        lcache_from = [False, True]
        ldatacol = [False, True]

        # Search by searchstr
        llcomb = [
            llambmin, llambmax, lion,
            lcache_from,
            ldatacol,
        ]
        ii, itot = -1, 2*2*3*2*2
        for comb in itt.product(*llcomb):
            ii += 1
            if all([vv is None for vv in comb[:2]]):
                continue
            if comb[2] == 'H' and comb[1] is None:
                continue
            if comb[2] == 'H' and all([vv is not None for vv in comb[:2]]):
                continue
            if any([vv is None for vv in comb[:2]]) and comb[2] != 'H':
                continue
            print(f'{ii} / {itot}  -  {comb}')

            # out = tfn.step01_search_online_by_wavelengthA(
                # lambmin=comb[0],
                # lambmax=comb[1],
                # ion=comb[2],
                # verb=True,
                # return_dout=True,
                # return_dsources=True,
                # cache_from=comb[3],
                # cache_info=True,
                # format_for_DataStock=comb[4],
                # create_custom=True,
            # )
            # del out

    def test02_clear_cache(self):
        tfn.clear_cache()



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 12 10:17:01 2024

@author: dvezinet
"""


import numpy as np
from matplotlib import path as mpath
import datastock as ds


from ..geom import _GG


# ###############################################################
# ###############################################################
#                      Main
# ###############################################################


def main(
    coll=None,
    key=None,
    strategy=None,
    nrays=None,
    # storing
    store=None,
    config=None,
    overwrite=None,
):

    # -------------
    # check
    # -------------

    key, strategy, nrays, store, overwrite = _check(
        coll=coll,
        key=key,
        strategy=strategy,
        nrays=nrays,
        # storing
        store=store,
        overwrite=overwrite,
    )

    # ---------------
    #  prepare
    # ---------------

    wdiag = 'diagnostic'
    key_cam = coll.dobj[wdiag][key]['camera']
    doptics = coll.dobj[wdiag][key]['doptics']

    # -------------------------
    # trivial: nrays=1 => LOS
    # -------------------------

    if nrays == 1:

        dout = {
            kcam: {}
            for kcam in key_cam
        }

        store = False

    # ---------------
    #  compute
    # ---------------

    # initialize
    dout = {}

    # loop on cameras
    for kcam in key_cam:

        # --------------
        # prepare pixel outline

        out0, out1 = coll.dobj['camera'][kcam]['dgeom']['outline']
        out0 = coll.ddata[out0]['data']
        out1 = coll.ddata[out1]['data']
        out_arr = np.array([out0, out1]).T

        # ---------------
        # call routine

        if strategy == 'random':

            dout[kcam] = _random(
                coll=coll,
                kcam=kcam,
                doptics=doptics[kcam],
                out_arr=out_arr,
                nrays=nrays,
                strategy=strategy,
            )

        elif strategy == 'outline':

            dout[kcam] = _outline()

        elif strategy == 'mesh':

            dout[kcam] = _mesh()

    # ---------------
    #  store
    # ---------------

    if store is True:
        _store(
            coll=coll,
            dout=dout,
            config=config,
            overwrite=overwrite,
        )

    else:
        return dout


# ###############################################################
# ###############################################################
#                      check
# ###############################################################


def _check(
    coll=None,
    key=None,
    strategy=None,
    nrays=None,
    # storing
    store=None,
    overwrite=None,
):

    # ------------
    # key
    # ------------

    wdiag = 'diagnostic'
    lok = list(coll.dobj.get(wdiag, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # spectro ?
    spectro = coll.dobj[wdiag][key]['spectro']
    if spectro is True:
        raise NotImplementedError()

    # ------------
    # strategy
    # ------------

    strategy = ds._generic_check._check_var(
        strategy, 'strategy',
        types=str,
        default='random',
        allowed=['random', 'mesh', 'outline'],
    )

    # ------------
    # nrays
    # ------------

    if strategy == 'custom':
        nrdef = 10
    else:
        nrdef = 10


    nrays = int(ds._generic_check._check_var(
        nrays, 'nrays',
        types=(int, float),
        default=nrdef,
        sign='>0',
    ))

    # ------------
    # store
    # ------------

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=True,
    )

    # ------------
    # overwrite
    # ------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    return key, strategy, nrays, store, overwrite


# ###############################################################
# ###############################################################
#                      random
# ###############################################################


def _random(
    coll=None,
    kcam=None,
    doptics=None,
    out_arr=None,
    nrays=None,
    strategy=None,
):

    # ---------------
    # prepare pixel
    # ---------------

    # outline path
    path = mpath.Path(out_arr)

    min0, max0 = np.min(out_arr[:, 0]), np.max(out_arr[:, 0])
    min1, max1 = np.min(out_arr[:, 1]), np.max(out_arr[:, 1])

    # areas to get safety factor
    area = coll.dobj['camera'][kcam]['dgeom']['pix_area']
    area_max = (max0 - min0) * (max1 - min1)

    # start point coords in pixel's plane
    start0, start1 = _seed_pix(
        path=path,
        nrays=nrays,
        nrays_factor=int(np.ceil(area_max/area) + 2),
        min0=min0,
        max0=max0,
        min1=min1,
        max1=max1,
    )

    # shared apertures ?
    pinhole = doptics['pinhole']
    parallel = coll.dobj['camera'][kcam]['dgeom']['parallel']

    # shape camera
    shape_cam = coll.dobj['camera'][kcam]['dgeom']['shape']

    # camera vector
    # get camera vectors
    dvect = coll.get_camera_unit_vectors(kcam)
    lc = ['x', 'y', 'z']
    if parallel is True:
        e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"] for kk in lc]
        e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"] for kk in lc]

    # -----------------------------------
    # prepare output
    # -----------------------------------

    shape_out = shape_cam + (nrays*nrays,)

    dout = {
        'key': f'{kcam}_rays_{strategy}',
        'start_x': np.full(shape_out, np.nan),
        'start_y': np.full(shape_out, np.nan),
        'start_z': np.full(shape_out, np.nan),
        'vect_x': np.full(shape_out, np.nan),
        'vect_y': np.full(shape_out, np.nan),
        'vect_z': np.full(shape_out, np.nan),
    }

    # -----------------------------------
    # pinhole camera (shared apertures)
    # -----------------------------------

    if pinhole is True:

        # get pixel centers
        cx, cy, cz = coll.get_camera_cents_xyz(kcam)
        cent_cam = np.r_[np.mean(cx), np.mean(cy), np.mean(cz)]

        # get end points
        end0, end1, op_cent, op_e0, op_e1 = _seed_optics(
            coll=coll,
            cent_cam=cent_cam,
            nrays=nrays,
            optics=doptics['optics'],
        )

        # full versions
        end0f = np.tile(end0, nrays)
        end1f = np.tile(end1, nrays)

        # full versions
        start0f = np.repeat(start0, end0.size)
        start1f = np.repeat(start1, end0.size)

        # vignetting polygons
        vignett_poly = [
            np.array(coll.get_optics_poly(k0, closed=True, add_points=False))
            for k0 in doptics['optics']
        ]
        lnvert = np.array([vv.shape[1] for vv in vignett_poly], dtype=np.int64)

        # loop on pixels
        ray_orig = np.full((3, nrays*end0.size), np.nan)
        ray_vdir = np.full((3, nrays*end0.size), np.nan)
        for ind in np.ndindex(shape_cam):

            sli_out = ind + (slice(None),)

            # unit vectors
            if parallel is not True:
                e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"][ind] for kk in lc]
                e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"][ind] for kk in lc]

            # ray_orig
            ray_orig[0, :] = cx[ind] + start0f * e0i_x + start1f * e1i_x
            ray_orig[1, :] = cy[ind] + start0f * e0i_y + start1f * e1i_y
            ray_orig[2, :] = cz[ind] + start0f * e0i_z + start1f * e1i_z

            # end points
            ray_vdir[0, :] = op_cent[0] + end0f * op_e0[0] + end1f * op_e1[0]
            ray_vdir[1, :] = op_cent[1] + end0f * op_e0[1] + end1f * op_e1[1]
            ray_vdir[2, :] = op_cent[2] + end0f * op_e0[2] + end1f * op_e1[2]

            # ray_vdir, normalized
            ray_vdir[0, :] = ray_vdir[0, :] - ray_orig[0, :]
            ray_vdir[1, :] = ray_vdir[1, :] - ray_orig[1, :]
            ray_vdir[2, :] = ray_vdir[2, :] - ray_orig[2, :]

            ray_norm = np.sqrt(np.sum(ray_vdir**2, axis=0))
            ray_vdir[:] = ray_vdir / ray_norm

            # vignetting (npoly, nrays) is wrong !!!
            # iok = _GG.vignetting(
            #     ray_orig,
            #     ray_vdir,
            #     vignett_poly,
            #     lnvert,
            #     num_threads=16,
            # )
            # print(ind, iok.sum(), iok.size)

            indrand = (np.random.randint(0, ray_orig.shape[1], nrays*nrays),)

            dout['start_x'][sli_out] = ray_orig[(0,) + indrand]
            dout['start_y'][sli_out] = ray_orig[(1,) + indrand]
            dout['start_z'][sli_out] = ray_orig[(2,) + indrand]
            dout['vect_x'][sli_out] = ray_vdir[(0,) + indrand]
            dout['vect_y'][sli_out] = ray_vdir[(1,) + indrand]
            dout['vect_z'][sli_out] = ray_vdir[(2,) + indrand]

    # ---------------
    # mesh pixels
    # ---------------

    else:

        for ind in np.ndindex(shape_cam):

            print(ind)


    return dout


# #########################
# seeding pixels
# #########################


def _seed_pix(
    path=None,
    nrays=None,
    nrays_factor=None,
    min0=None,
    max0=None,
    min1=None,
    max1=None,
):

    seed0 = np.random.random((nrays*nrays_factor,))
    seed1 = np.random.random((nrays*nrays_factor,))

    pts0 = (max0 - min0) * seed0 + min0
    pts1 = (max1 - min1) * seed1 + min1

    # only keep those in the pixel
    iok_pix = path.contains_points(np.array([pts0, pts1]).T).nonzero()[0]

    # imax
    imax = min(nrays, iok_pix.size)

    return pts0[iok_pix[:imax]], pts1[iok_pix[:imax]]


# #########################
# seeding optics
# #########################


def _seed_optics(
    coll=None,
    cent_cam=None,
    nrays=None,
    optics=None,
):

    # -----------------------------
    # select smallest etendue optic
    # -----------------------------

    optics, optics_cls = coll.get_optics_cls(optics)

    etend = [
        (
            coll.dobj[cc][kop]['dgeom']['area']
            / np.linalg.norm(coll.dobj[cc][kop]['dgeom']['cent'] - cent_cam)**2
        )
        for kop, cc in zip(optics, optics_cls)
    ]

    ind = np.nanargmin(etend)
    kop = optics[ind]
    cc = optics_cls[ind]

    # -----------------------------
    # seed
    # -----------------------------

    out0, out1 = coll.get_optics_outline(kop, add_points=False, closed=False)

    path = mpath.Path(np.array([out0, out1]).T)

    min0, max0 = np.min(out0), np.max(out0)
    min1, max1 = np.min(out1), np.max(out1)

    area = coll.dobj[cc][kop]['dgeom']['area']
    factor = int(np.ceil((max0 - min0) * (max1 - min1) / area) + 2)

    seed0 = np.random.random((nrays * factor,))
    seed1 = np.random.random((nrays * factor,))

    pts0 = (max0 - min0) * seed0 + min0
    pts1 = (max1 - min1) * seed1 + min1

    # only keep those in the pixel
    iok_pix = path.contains_points(np.array([pts0, pts1]).T).nonzero()[0]

    # imax
    imax = min(nrays*2, iok_pix.size)

    # optic cent and vectors
    cent = coll.dobj[cc][kop]['dgeom']['cent']
    e0 = coll.dobj[cc][kop]['dgeom']['e0']
    e1 = coll.dobj[cc][kop]['dgeom']['e1']

    return pts0[iok_pix[:imax]], pts1[iok_pix[:imax]], cent, e0, e1


# ###############################################################
# ###############################################################
#                      outline
# ###############################################################


def _outline():


    return


# ###############################################################
# ###############################################################
#                      mesh
# ###############################################################


def _mesh():


    return


# ###############################################################
# ###############################################################
#                      store
# ###############################################################


def _store(
    coll=None,
    kdiag=None,
    dout=None,
    config=None,
    overwrite=None,
):

    # -----------------
    # add ref
    # -----------------

    nrays = list(dout.values())[0]['start_x'].shape[-1]

    kref = f"{kdiag}_nrays"
    coll.add_ref(key=kref, size=nrays)

    # --------------
    # add rays
    # --------------

    for kcam, v0 in dout.items():

        ref = coll.dobj['camera'][kcam]['dgeom']['ref'] + (kref,)
        coll.add_rays(
            ref=ref,
            config=config,
            **v0
        )

    return


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True
#

ctypedef struct ChainedList:
    double val
    int size
    ChainedList* next
    ChainedList* last

cdef ChainedList* create_ordered(int N) nogil
cdef void pop_at_pos(ChainedList** head, int n) nogil
cdef double get_at_pos(ChainedList* head, int n) nogil
cdef void push_back(ChainedList** head, double val) nogil
cdef void free_cl(ChainedList** head) nogil



# -*- coding: utf-8 -*-
"""
The nist-compatibility module of tofu

"""
import traceback

try:
    try:
        from tofu.nist2tofu._requests import *
    except Exception:
        from ._requests import *
except Exception as err:
    msg = str(traceback.format_exc())
    msg += "\n\n    => the optional sub-package tofu.nist2tofu is not usable\n"
    raise Exception(msg)

del traceback
